from typing import Any, Callable, NamedTuple
from redis import Redis
from redis.commands.core import Script
from .constants import INCREMENT_LUA_SCRIPT


class Subscriber(NamedTuple):
    redis: Redis
    score_key: str
    """
    The Redis key for the score.
    """
    score_member: str
    """
    The Redis member value for the score.
    """
    default_score: float
    """
    The default score.
    """
    min_score: float
    """
    The minimum score.
    """
    max_score: float
    """
    The maximum score.
    """
    handler: Callable[["Subscriber", Any], None]  # type: ignore
    increment_script: Script

    @staticmethod
    def new(
        redis: Redis,
        score_key: str = None,  # type: ignore
        score_member: str = None,  # type: ignore
        default_score: float = 0,
        min_score: float = 1,
        max_score: float = 100,
        handler: Callable[["Subscriber", Any], None] = None,  # type: ignore
    ):
        return Subscriber(
            redis=redis,
            score_key=score_key,
            score_member=score_member,
            default_score=default_score,
            min_score=min_score,
            max_score=max_score,
            handler=handler,
            increment_script=redis.register_script(
                INCREMENT_LUA_SCRIPT,
            ),  # type: ignore
        )

    @property
    def score(self) -> float:
        score = self.redis.zscore(
            self.score_key,
            self.score_member,
        )

        if score == None:
            self.redis.zadd(
                self.score_key,
                {
                    self.score_member: self.default_score,
                },
                nx=True,
            )

            return self.default_score

        return score  # type: ignore

    def add_score(self, value: float):
        return self.increment_script(
            (self.score_key,),
            (
                self.score_member,
                value,
                self.default_score,
                self.min_score,
                self.max_score,
            ),
        )

    def set_score(self, value: float):
        self.redis.zadd(
            self.score_key,
            {
                self.score_member: value,
            },
            xx=True,
        )

        return value

    def reset_score(self):
        self.set_score(self.default_score)

        return self.default_score



from typing import Iterable

from redis_director.publisher import Publisher


def generic_sadd_callback(
    publisher: Publisher,
    payloads: Iterable[str],
):
    return publisher.redis.sadd(
        publisher.queue_key,
        *payloads,
    )


def generic_rpush_callback(
    publisher: Publisher,
    payloads: Iterable[str],
):
    return publisher.redis.rpush(
        publisher.queue_key,
        *payloads,
    )


def generic_spop_callback(
    publisher: Publisher,
    batch_size: int,
):
    return publisher.redis.spop(
        publisher.queue_key,
        batch_size,
    )


def generic_lpop_callback(
    publisher: Publisher,
    batch_size: int,
):
    return publisher.redis.lpop(
        publisher.queue_key,
        batch_size,
    )



from .handler import generic_list_handler, generic_set_handler
from .publisher import Publisher
from .subscriber import Subscriber
from .callback import (
    generic_lpop_callback,
    generic_spop_callback,
    generic_rpush_callback,
    generic_sadd_callback,
)



from .subscriber import Subscriber


def generic_set_handler(
    key: str,
    add_score: float = -1,
):
    def handler(route: Subscriber, payload):
        route.add_score(add_score)
        route.redis.sadd(key, payload)

    return handler


def generic_list_handler(
    key: str,
    add_score: float = -1,
):
    def handler(route: Subscriber, payload):
        route.add_score(add_score)
        route.redis.lpush(key, payload)

    return handler



INCREMENT_LUA_SCRIPT = """
local key = KEYS[1]
local member = ARGV[1]
local increment = tonumber(ARGV[2])
local minScore = tonumber(ARGV[4])
local maxScore = tonumber(ARGV[5])

local currentScore = redis.call('zscore', key, member)

if not currentScore then
    currentScore = tonumber(ARGV[3])
else
    currentScore = tonumber(currentScore)
end

local newScore = currentScore + increment

if newScore < minScore then
    newScore = minScore

elseif newScore > maxScore then
    newScore = maxScore

end

redis.call('zadd', key, newScore, member)

return newScore
"""



from typing import Any, Callable, Dict, Iterable, List
from fun_things.math import weighted_distribution
from redis import Redis

from .subscriber import Subscriber


class Publisher:
    def __init__(
        self,
        redis: Redis,
        score_key: str,
        queue_key: str,
        add_payloads_callback: Callable[
            [
                "Publisher",
                Iterable[str],
            ],
            None,
        ] = None,  # type: ignore
        pop_payloads_callback: Callable[
            [
                "Publisher",
                int,
            ],
            List[str],
        ] = None,  # type: ignore
    ):
        self.__subscribers: Dict[str, Subscriber] = {}

        self.__redis = redis
        self.__score_key: str = score_key
        self.__queue_key: str = queue_key
        self.__add_payloads_callback = add_payloads_callback
        self.__pop_payloads_callback = pop_payloads_callback

        from redis_director.callback import (
            generic_sadd_callback,
            generic_spop_callback,
        )

        if add_payloads_callback == None:
            self.__add_payloads_callback = generic_sadd_callback

        if pop_payloads_callback == None:
            self.__pop_payloads_callback = generic_spop_callback

    @property
    def redis(self):
        return self.__redis

    @property
    def score_key(self):
        return self.__score_key

    @property
    def queue_key(self):
        return self.__queue_key

    def reset_scores(self):
        for subscriber in self.__subscribers.values():
            subscriber.reset_score()

    def add_subscriber(
        self,
        score_member: str,
        handler: Callable[[Subscriber, Any], None],
        default_score: float = 100,
        min_score: float = 1,
        max_score: float = 100,
    ):
        return self.add_subscribers(
            Subscriber.new(
                redis=self.__redis,
                score_key=self.__score_key,
                score_member=score_member,
                default_score=default_score,
                min_score=min_score,
                max_score=max_score,
                handler=handler,
            )
        )

    def add_subscribers(self, *subscribers: Subscriber):
        for subscriber in subscribers:
            self.__subscribers[subscriber.score_member] = subscriber

        return self

    def get_subscriber(self, member: str) -> Subscriber:
        if member not in self.__subscribers:
            return None  # type: ignore

        return self.__subscribers[member]

    @property
    def random_subscriber(self):
        """
        Returns 1 random subscriber
        based on their scores.

        The higher score has the higher chance
        of being chosen.
        """
        subscribers = weighted_distribution(
            self.__subscribers.values(),
            lambda subscriber: subscriber.score,
        )

        for subscriber in subscribers:
            return subscriber

        raise ValueError("No subscribers found!")

    @property
    def random_subscribers(self):
        """
        Infinitely returns a random subscriber
        based on their scores.

        The higher score has the higher chance
        of being chosen.
        """
        while True:
            yield self.random_subscriber

    def add_payloads(self, *payloads: str):
        if self.__add_payloads_callback == None:
            raise TypeError("`add_payloads_callback` not provided!")

        self.__add_payloads_callback(
            self,
            payloads,
        )

        return self

    def publish(
        self,
        batch_size=1,
    ):
        if self.__pop_payloads_callback == None:
            raise TypeError("`pop_payloads_callback` not provided!")

        payloads = self.__pop_payloads_callback(
            self,
            batch_size,
        )

        if payloads == None:
            return self

        for payload in payloads:
            for subscriber in self.random_subscribers:
                if subscriber.handler != None:
                    subscriber.handler(subscriber, payload)

                break

        return self
